//noinspection ThisExpressionReferencesGlobalObjectJS
(function (host) {

    var URL = function (url) {
        if (url) {
            if (typeof url !== "string") {
                throw "url parameter should be string";
            }
            this.parse(url);
        }
    };

    URL.prototype.parse = function (url) {
        var urlSlicer = new URLSlicer(url);
        this.parseHref(urlSlicer);
        this.parseProtocol(urlSlicer);
        this.parseSlashes(urlSlicer);
        this.parseHost(urlSlicer);
        this.parsePathname(urlSlicer);
        this.parseSearch(urlSlicer);
        this.parseHash(urlSlicer);
        deleteUndefinedProperty(this);
    };
        URL.prototype.parseHref = function (urlSlicer) {
            this.href = urlSlicer.urlFragment;
        };
        URL.prototype.parseProtocol = function (urlSlicer) {
            this.protocol = toLowerCase(urlSlicer.slice(/^([a-z0-9.+-]+:)/i));
        };
        URL.prototype.parseSlashes = function (urlSlicer) {
            this.slashes = !!urlSlicer.slice(/^(\/\/)/);
        };
        URL.prototype.parseHost = function (urlSlicer) {
            if (!this.isHostlessProtocol()) {
                // hostname should be started with an alphabet or a digit
                var rawHost = urlSlicer.slice(/^([a-zA-Z0-9][^\/;\?#]+)/);
                if (rawHost) {
                    var indexOfAuthDelimeter = rawHost.indexOf("@");
                    this.auth = rawHost.substring(0, indexOfAuthDelimeter);
                    this.hostname = rawHost.substring(indexOfAuthDelimeter + 1).toLowerCase();

                    this.host = this.auth + (this.auth ? "@" : "") + this.hostname;
                }
            }
        };
        URL.prototype.parsePathname = function (urlSlicer) {
            this.pathname = urlSlicer.slice(/([^\?#]+)/) || "/";
        };
        URL.prototype.parseSearch = function (urlSlicer) {
            this.search = urlSlicer.slice(/(\?[^#]+)/);
            if (this.search) {
                var query = this.search.substring(1);
                this.query = query;
                this.queryObject = parseQuery(query);
            }
        };
        URL.prototype.parseHash = function (urlSlicer) {
            this.hash = urlSlicer.slice(/(#.+)/);
        };
        URL.prototype.isHostlessProtocol = function () {
            return this.protocol === "javascript:";
        };

    URL.prototype.format = function () {
        var buf = [];
        // protocol
        buf.push(this.protocol);
        if (!this.protocol.match(/:$/)) {
            buf.push(':');
        }
        // slashes
        if (isSlashedProtocol(this.protocol)) {
            buf.push('//');
        }

        // host
        if (this.host) {
            buf.push(this.host);
        } else {
            if (this.auth) {
                buf.push(this.auth);
                buf.push('@');
            }
            buf.push(this.hostname);
            if (this.port) {
                buf.push(':');
                buf.push(this.port);
            }
        }

        // path
        if (this.pathname) {
            if (!this.pathname.match(/^\//)) {
                buf.push('/');
            }
            buf.push(this.pathname);
        }

        // search (query)
        if (this.search) {
            if (!this.search.match(/^\?/)) {
                buf.push('?');
            }
            buf.push(this.search);
        } else if (this.query) {
            buf.push('?');
            buf.push(this.query);
        } else if (this.queryObject) {
            buf.push('?');
            buf.push(formatQueryObject(this.queryObject));
        }

        // hash
        if (this.hash) {
            if (!this.hash.match(/^#/)) {
                buf.push('#');
            }
            buf.push(this.hash);
        }

        return buf.join('');
    };

    URL.prototype.resolve = function (pathname) {
        var newPathname;
        if (pathname.charAt(0) === '/') {
            newPathname = pathname;
        } else {
            var basepath = this.pathname.substr(0, this.pathname.lastIndexOf('/'));
            newPathname = basepath + '/' + pathname;
        }
        this.pathname = this.resolvePathname(newPathname);
        return this.format();
    };

    // TODO rename it
    URL.prototype.resolvePathname = function (pathname) {
        var pathElements = pathname.split('/'),
            resolvedPathStack = [];

        for (var i = 0; i < pathElements.length; i++) {
            if (pathElements[i] === '..') {
                if (resolvedPathStack[resolvedPathStack.length - 1] !== '') {
                    resolvedPathStack.pop();
                }
            } else if (pathElements[i] === '.') {
            } else {
                resolvedPathStack.push(pathElements[i]);
            }
        }
        return resolvedPathStack.join('/');
    };


    URL.parse = function (urlString) {
        return new URL(urlString);
    };

    URL.format = function(urlObject) {
        return URL.prototype.format.call(urlObject);
    };

    URL.resolve = function(baseUrl, targetUrl) {
        return new URL(baseUrl).resolve(targetUrl);
    };

    var URLSlicer = function (url) {
        this.urlFragment = url;
    };
    URLSlicer.prototype = {
        slice: function (pattern) {
            var matches = this.urlFragment.match(pattern);
            if (matches && matches.length == 2) {
                this.urlFragment = this.urlFragment.substring(matches[1].length);
                return matches[1];
            }
            return undefined;
        }
    };


    var SLASHED_PROTOCOLS = ['http', 'https', 'ftp', 'gopher', 'file'];
    function isSlashedProtocol(protocol) {
        protocol = protocol.replace(/:$/, '');
        for (var i = 0; i < SLASHED_PROTOCOLS.length; i++) {
            if (protocol === SLASHED_PROTOCOLS[i]) {
                return true;
            }
        }
        return false;
    }

    function formatQueryObject(queryObject) {
        var query = [];
        for (var key in queryObject) if (queryObject.hasOwnProperty(key)) {
            query.push(key + '=' + queryObject[key]);
        }
        return query.join('&');
    }

    function parseQuery(query) {
        var queryObject = {};
        var parameters = query.split("&");
        for (var i = 0; i < parameters.length; i++) {
            var kvp = parameters[i].split("=");
            //noinspection UnnecessaryLocalVariableJS
            var key = kvp[0], value = kvp[1];
            queryObject[key] = value;
        }
        return queryObject;
    }

    // helper functions
    function deleteUndefinedProperty(obj) {
        for (var key in obj)
            if (obj.hasOwnProperty(key)) {
                if (typeof obj[key] == "undefined") {
                    delete obj[key];
                }
            }
    }

    function toLowerCase(value) {
        return value ? value.toLowerCase() : value;
    }

    host.URL = URL;
})(this);