describe("URL interface", function () {
    it("should define URL class", function () {
        expect(URL).toBeDefined();
    });

    it("should have url constructor parameter", function () {
        expect(
            function () {
                new URL();
            }).not.toThrow("url parameter is missed");
    });

    it("url parameter should be string", function () {
        expect(
            function () {
                new URL({});
            }).toThrow("url parameter should be string");
    });
});

describe("URL parsing", function () {
    it("should keep the url as href", function () {
        var url = "http://www.example.com/";
        expect(new URL(url).href).toBe(url);
    });

    it("should parse protocol", function () {
        expect(new URL("http://www.example.com/").protocol).toBe("http:");
        expect(new URL("https://www.example.com/").protocol).toBe("https:");
        expect(new URL("ftp://file.example.com/").protocol).toBe("ftp:");
    });

    it("protocol should be lowercased", function () {
        expect(new URL("HTTP://www.example.com/").protocol).toBe("http:");
    });

    it("should parse slashes", function () {
        expect(new URL("http://www.example.com/").slashes).toBeTruthy();
    });

    it("should parse host from simple url", function () {
        var urlSimple = new URL("http://www.example.com/");
        expect(urlSimple.host).toBe("www.example.com");
    });

    it("should parse host from url with auth", function () {
        var urlWithAuth = new URL("http://user:pass@www.example.com/");
        expect(urlWithAuth.host).toBe("user:pass@www.example.com");
    });

    it("hostname part only should be lowercased, but auth part should not", function () {
        expect(new URL("http://User:Pass@WWW.EXAMPLE.COM/").host).toBe("User:Pass@www.example.com");
    });

    it("should parse hostname from simple url", function () {
        var urlSimple = new URL("http://www.example.com/");
        expect(urlSimple.hostname).toBe("www.example.com");
    });

    it("hostname should be lowercased", function () {
        expect(new URL("http://WWW.EXAMPLE.COM/").hostname).toBe("www.example.com");
    });

    it("should parse hostname from url with auth", function () {
        var urlWithAuth = new URL("http://user:pass@www.example.com/");
        expect(urlWithAuth.hostname).toBe("www.example.com");
    });

    it("should parse auth", function () {
        var urlWithAuth = new URL("http://user:pass@www.example.com/");
        expect(urlWithAuth.auth).toBe("user:pass");
    });

    it("should parse simple path", function () {
        var url = new URL("http://cafe.example.com/ilovenba");
        expect(url.pathname).toBe("/ilovenba");
    });

    it("should parse multi level path", function () {
        var url = new URL("http://cafe.example.com/ilovenba/asdf/39248");
        expect(url.pathname).toBe("/ilovenba/asdf/39248");
    });

    it("should not parse hash as pathname", function () {
        var url = new URL("http://cafe.example.com/ilovenba/asdf/39248#clicked");
        expect(url.pathname).toBe("/ilovenba/asdf/39248");
    });

    it("should not parse query as pathname", function () {
        var url = new URL("http://cafe.example.com/ilovenba/asdf/39248?q=foo");
        expect(url.pathname).toBe("/ilovenba/asdf/39248");
    });

    it("should assign '/' (slash) as default pathname", function () {
        var url = new URL("http://www.example.com");
        expect(url.pathname).toBe("/");
    });

    it("should parse search", function () {
        var urlWithSearch = new URL("http://www.example.com/page?foo=bar");
        expect(urlWithSearch.search).toBe("?foo=bar");
    });

    it("should parse search from url with two parameter", function () {
        var urlWithSearch = new URL("http://www.example.com/page?foo=bar&foo1=bar1");
        expect(urlWithSearch.search).toBe("?foo=bar&foo1=bar1");
    });

    it("should parse search from url with hash", function () {
        var urlWithSearch = new URL("http://www.example.com/page?foo=bar#hash");
        expect(urlWithSearch.search).toBe("?foo=bar");
    });

    it("should parse query", function () {
        var urlWithSearch = new URL("http://www.example.com/page?foo=bar");
        expect(urlWithSearch.query).toBe("foo=bar");
    });

    it("should parse query as object", function () {
        var urlWithSearch = new URL("http://www.example.com/page?foo=bar&p1=v1");
        expect(urlWithSearch.queryObject.foo).toBe("bar");
        expect(urlWithSearch.queryObject.p1).toBe("v1");
    });

    it("should parse hash", function () {
        var urlWithHash = new URL("http://www.example.com/#hash");
        expect(urlWithHash.hash).toBe("#hash");
    });

    beforeEach(function () {
        this.addMatchers({
            toEqualFieldsOnly: function (expected) {
                var equal = true;
                for (var key in this.actual)
                    if (this.actual.hasOwnProperty(key)) {
                        if (typeof this.actual[key] != "function") {
                            equal = equal && (this.actual[key] == expected[key]);
                        }
                    }
                return equal;
            }
        });
    });

    // see http://en.wikipedia.org/wiki/File_URI_scheme
    it("should parse local file path", function () {
        expect(new URL("file:///Users/guest")).toEqualFieldsOnly({
            protocol: "file:",
            slashes: true,
            pathname: "/Users/guest",
            href: "file:///Users/guest"
        });
    });

    it("should parse host from file path", function () {
        expect(new URL("file://localhost/Users/guest").host).toEqual("localhost");
    });

    it("should parse windows style file path", function () {
        expect(new URL("file:///c:/WINDOWS/clock.avi")).toEqualFieldsOnly({
            protocol: "file:",
            slashes: true,
            pathname: "/c:/WINDOWS/clock.avi",
            href: "file:///c:/WINDOWS/clock.avi"
        });
    });

    it("should parse path only url", function () {
        expect(new URL("/some/path/only.html").pathname).toBe("/some/path/only.html");
    });

    it("should parse hostless and protocoless url", function () {
        var url = new URL("/some/path/only.html?param1=foo#hash");
        expect(url.pathname).toBe('/some/path/only.html');
        expect(url.hash).toBe('#hash');
        expect(url.slashes).toBeFalsy();
        expect(url.query).toBe('param1=foo');
        expect(url.hash).toBe('#hash');
    });

    it("should parse relative url (../ or ./)", function () {
        expect(new URL("../relative/path.html").pathname).toBe("../relative/path.html");
        expect(new URL("./relative/path.html").pathname).toBe("./relative/path.html");
    });

    it("should parse relative part in pathname", function () {
        expect(new URL("http://www.example.com/some/path/../relative/path.html").pathname)
            .toBe("/some/path/../relative/path.html");
    });

    it("shoud absolutize relative pathname");

    it("should support hostless protocol", function () {
        var url = new URL("javascript:void(0);");
        expect(url.protocol).toBe("javascript:");
        expect(url.pathname).toBe("void(0);");
    });
});

describe('URL formatting', function () {
    var url;
    beforeEach(function () {
        url = new URL();
        url.hostname = 'www.example.com';
    });

    it('href will be ignored.');
    it('protocol should be treated the same with or without the trailing : (colon).', function () {
        url.protocol = 'ftp:';
        expect(url.format()).toBe('ftp://www.example.com');
        url.protocol = 'ftp';
        expect(url.format()).toBe('ftp://www.example.com');
    });

    it('The protocols http, https, ftp, gopher, file will be postfixed with :// (colon-slash-slash).', function () {
        function shouldBePostfixedWithColonSlashesIfProtocolIs(protocol) {
            url.protocol = protocol;
            expect(url.format()).toBe(protocol + '://www.example.com');
        }

        shouldBePostfixedWithColonSlashesIfProtocolIs('http');
        shouldBePostfixedWithColonSlashesIfProtocolIs('https');
        shouldBePostfixedWithColonSlashesIfProtocolIs('ftp');
        shouldBePostfixedWithColonSlashesIfProtocolIs('gopher');
        shouldBePostfixedWithColonSlashesIfProtocolIs('file');
    });

    it('All other protocols mailto, xmpp, aim, sftp, foo, etc will be postfixed with : (colon)', function () {
        function shouldNotBePostfixedWithColonSlashesIfProtocolIs(protocol) {
            url.protocol = protocol;
            expect(url.format()).toBe(protocol + ':www.example.com');
        }

        shouldNotBePostfixedWithColonSlashesIfProtocolIs('mailto');
        shouldNotBePostfixedWithColonSlashesIfProtocolIs('xmpp');
        shouldNotBePostfixedWithColonSlashesIfProtocolIs('sftp');
    });

    it('auth, hostname and port will be used if host is absent', function () {
        url.protocol = 'http:';
        url.hostname = 'hostname';
        url.auth = 'user:pass';
        url.port = '80';
        expect(url.format()).toBe('http://user:pass@hostname:80');
    });

    it('auth will only be used if host is absent.', function () {
        url.protocol = 'http:';
        url.host = 'user:pass@hostname';
        url.auth = 'anotheruser:pass';
        expect(url.format()).toBe('http://user:pass@hostname');
    });

    it('hostname will only be used if host is absent.', function () {
        url.protocol = 'http:';
        url.host = 'hostname';
        url.hostname = 'anotherhostname';
        expect(url.format()).toBe('http://hostname');
    });

    it('port will only be used if host is absent.', function () {
        url.protocol = 'http:';
        url.host = 'hostname:80';
        url.port = '8080';
        expect(url.format()).toBe('http://hostname:80');
    });

    it('pathname is treated the same with or without the leading / (slash)', function () {
        url.protocol = 'http:';
        url.pathname = '/path/to/page.html';
        expect(url.format()).toBe('http://www.example.com/path/to/page.html');
        url.pathname = 'path/to/page.html';
        expect(url.format()).toBe('http://www.example.com/path/to/page.html');
    });

    it('search will be used in place of query', function () {
        url.protocol = 'http:';
        url.search = '?foo=search';
        url.query = 'foo=query';
        expect(url.format()).toBe('http://www.example.com?foo=search');
    });

    it('query will only be used if search is absent.', function () {
        url.protocol = 'http:';
        url.query = 'foo=query';
        expect(url.format()).toBe('http://www.example.com?foo=query');
    });

    it('queryObject will only be used if search and query is absent.', function () {
        url.protocol = 'http:';
        url.queryObject = { foo: 'queryObject', bar: 'bar' };
        expect(url.format()).toBe('http://www.example.com?foo=queryObject&bar=bar');
    });

    it('search is treated the same with or without the leading ? (question mark)', function () {
        url.protocol = 'http:';
        url.pathname = '/path';
        url.search = '?with=questionmark';
        expect(url.format()).toBe('http://www.example.com/path?with=questionmark');
        url.search = 'without=questionmark';
        expect(url.format()).toBe('http://www.example.com/path?without=questionmark');
    });

    it('hash is treated the same with or without the leading # (pound sign, anchor)', function () {
        url.protocol = 'http:';
        url.pathname = '/path';
        url.hash = '#with_anchor';
        expect(url.format()).toBe('http://www.example.com/path#with_anchor');
        url.hash = 'without_anchor';
        expect(url.format()).toBe('http://www.example.com/path#without_anchor');
    });
});

describe('URL.resolve', function () {

    describe("resolvePathname", function () {
        var resolvePath;

        beforeEach(function () {
            resolvePath = URL.prototype.resolvePathname;
        });

        it("should return normal pathnames", function () {
            expect(resolvePath('/path/to/page.html')).toBe('/path/to/page.html');
            expect(resolvePath('/path/to/')).toBe('/path/to/');
            expect(resolvePath('path/to/')).toBe('path/to/');
            expect(resolvePath('/')).toBe('/');
        });

        it("should resolve parent path ('../') path element", function () {
            expect(resolvePath('/path/../to/page.html')).toBe('/to/page.html');
            expect(resolvePath('/path/../to/../page.html')).toBe('/page.html');
        });

        it("should ignore parent path element as prefix", function () {
            expect(resolvePath('/../path/page.html')).toBe('/path/page.html');
            expect(resolvePath('/../../path/page.html')).toBe('/path/page.html');
        });

        it("should resolve current path ('./')element", function () {
            expect(resolvePath('/path/./to/page.html')).toBe('/path/to/page.html');
            expect(resolvePath('/path/./to/./page.html')).toBe('/path/to/page.html');
            expect(resolvePath('/./to/page.html')).toBe('/to/page.html');
            expect(resolvePath('./to/page.html')).toBe('to/page.html');
        });

        it('should resolve current path and parent path combined one', function () {
            expect(resolvePath('/path/./to/.././page.html')).toBe('/path/page.html');
        })
    });

    describe('url.resolve(to)', function () {
        it('should resolve relative path appending', function () {
            var url = new URL('http://www.example.com/base/path');
            expect(url.resolve('new/relative/page.html')).toBe('http://www.example.com/base/new/relative/page.html');

            url = new URL('http://www.example.com/base/path/');
            expect(url.resolve('new/relative/page.html')).toBe('http://www.example.com/base/path/new/relative/page.html');
        });

        it('should resolve simple path reset', function () {
            var url = new URL('http://www.example.com/base/path');
            expect(url.resolve('/new/reseted/page.html')).toBe('http://www.example.com/new/reseted/page.html');
        });

        it('should resolve relative path which include parent path reference', function () {
            var url = new URL('http://www.example.com/base/path');
            expect(url.resolve('../relative/page.html')).toBe('http://www.example.com/relative/page.html');
            url = new URL('http://www.example.com/base/path/');
            expect(url.resolve('../../relative/page.html')).toBe('http://www.example.com/relative/page.html');
        });

    });

    describe('url should provide static interfaces', function () {

        it('should provide static parse method', function () {
            var url = URL.parse('http://www.example.com/base/path');
            expect(url.protocol).toBe('http:');
            expect(url.host).toBe('www.example.com');
            expect(url.pathname).toBe('/base/path');
        });

        it('should provide static format method', function () {
            var url = {
                protocol: 'http:',
                host: 'www.example.com',
                pathname: '/base/path'
            };
            expect(URL.format(url)).toBe('http://www.example.com/base/path');
        });

        it('should provide static resolve method', function () {
            var baseUrl = 'http://www.example.com/base/path',
                targetUrl = '../relative/page.html';

            expect(URL.resolve(baseUrl, targetUrl)).toBe('http://www.example.com/relative/page.html');
        });
    });
});

// spec http://nodejs.org/docs/v0.6.0/api/url.html
// TODO handle exceptions

// snippet
// var url = require('url');